import os
import argparse
import yaml

from access_niu.pipeline import build_pipeline


def _create_parser():
    parser = argparse.ArgumentParser(description="access-niu parser")
    parser.add_argument("--template", type=str, required=True, help="Project template")

    return parser.parse_args()


class Trainer(object):
    """Trainer class
    """
    def __init__(self, template):
        self.template = template
        self.pipeline = []

    def start_construction(self):
        """Builds the pipeline from given template.
        """
        self.pipeline = build_pipeline(self.template)

    def train(self):
        """Trains the model using the constructed pipeline.
        """

        kwargs = self.template

        for component in self.pipeline:
            result = component.execute(**kwargs)
            if result is not None:
                kwargs.update(result)

    def persist(self):
        """ Saves the trained model.
        """

        kwargs = self.template
        for component in self.pipeline:
            component.persist(**kwargs)

        keys_to_delete = [k for k in kwargs.keys() if 'generator' in k]
        for k in keys_to_delete:
                del kwargs[k]

        with open(os.path.join(kwargs.get('project').get('path'), 'template.yml'), 'w') as f:
            yaml.safe_dump(kwargs, f)


if __name__ == "__main__":
    args = _create_parser()

    with open(args.template) as f:
        template = yaml.safe_load(f)

    trainer = Trainer(template)
    trainer.start_construction()
    trainer.train()
    trainer.persist()



from keras.preprocessing import image


def data_generator(path, img_ht, img_wt, batch_size=32):

    gen = image.ImageDataGenerator(rescale=1.0 / 255)

    generator = gen.flow_from_directory(
        path, target_size=(img_ht, img_wt), batch_size=batch_size, shuffle=True
    )

    labels = {v: k for k, v in generator.class_indices}

    return generator, labels, generator.n



import os

from keras.models import model_from_json
from keras.models import Model
import yaml


def load_model(model_dir: str) -> [Model, dict]:
    """Loads model from json and h5py file

    :param model_dir: Path to json model file
    :return: Keras model file
    """

    with open(os.path.join(model_dir, "model.json")) as f:
        model = model_from_json(f.read())

    weights_file = os.path.join(model_dir, "model_weights.h5")
    model.load_weights(weights_file)

    # look at: https://github.com/keras-team/keras/issues/6462
    model._make_predict_function()

    with open(os.path.join(model_dir, "labels.yml")) as f:
        labels = yaml.safe_load(f)

    return model, labels


def parse(model, data):
    pred = model.predict(data, verbose=0)
    return pred



from access_niu.components.component import ComponentManager

comp_manager = ComponentManager()


def _build_component(name, **kwargs):
    return comp_manager.get(name)(**kwargs)


def _get_data_component(template):
    components = []
    input_layer = template.get("pipeline")[0].get("input_layer")
    image_size = {
        "image_height": input_layer.get("image_height"),
        "image_width": input_layer.get("image_width"),
    }

    for dir_type, path in template.get("data").items():
        kwargs = {}
        kwargs["data_dir"] = path
        kwargs["generator_name"] = f"{dir_type}_generator"
        kwargs["num_sample_name"] = f"n_{dir_type}_samples"
        kwargs["batch_size"] = template.get("batch_size")
        kwargs.update(image_size)
        component = comp_manager.get("data_generator")()
        result = component.prepare(**kwargs)
        if result is not None and type(result) == dict:
            template.update(result)

        components.append(component)

    return components


def build_pipeline(template):

    pipeline = _get_data_component(template)

    for comp in template.get("pipeline"):
        pipeline.append(_build_component(**comp))

    return pipeline



import os

import yaml

from access_niu.utility import ensure_directory


def save_keras_model(path, model, labels):

    ensure_directory(path)

    with open(os.path.join(path, "model.json"), "w") as f:
        f.write(model.to_json())

    model.save(os.path.join(path, "model_weights.h5"))
    _save_labels(path, labels)


def _save_labels(path, labels):

    with open(os.path.join(path, "labels.yml"), "w") as f:
        yaml.safe_dump(labels, f)



from keras.layers import Dense, Input, GlobalAveragePooling2D
from keras.models import Model
from keras.applications.mobilenet_v2 import MobileNetV2


def get_model():

    input_tensor = Input(shape=(224, 224, 3))
    base_model = MobileNetV2(
        input_shape=(224, 224, 3),
        include_top=False,
        weights="imagenet",
        input_tensor=input_tensor,
    )
    x = base_model.output
    x = GlobalAveragePooling2D()(x)
    op = Dense(2, activation="softmax", name="final_output")(x)

    for layer in base_model.layers:
        layer.trainable = False

    model = Model(inputs=[input_tensor], outputs=[op])

    model.summary()

    return model



from setuptools import find_namespace_packages
from setuptools import setup
from access_niu.about import __version__, __license__

setup(
    name="access-niu",
    classifiers=[
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.6",
            "Operating System :: OS Independent",
            "Topic :: Software Development :: Libraries",
        ],
    author="ConvexHull Technology Private Limited",
    author_email="connect@accessai.co",
    maintainer="Majeed Khan",
    maintainer_email="majeed.khan@accessai.co",
    version=__version__,
    packages= find_namespace_packages(exclude=["sample", "tests"]),
    license=__license__,
    long_description=open("README.md").read(),
    install_requires=[
        "keras==2.2.4",
        "numpy==1.16.2",
        "scikit-learn==0.20.3",
        "tensorflow==1.13.1",
        "pillow==6.0.0",
        "h5py==2.9.0",
        "black",
        "gunicorn==19.9.0",
        "flask==1.0.2",
    ],
)



import abc
from abc import ABCMeta
from collections import defaultdict

from keras.applications.mobilenet_v2 import MobileNetV2
from keras.applications.mobilenet import MobileNet
from keras.layers import Dense, Input, GlobalAveragePooling2D
from keras.models import Model
from keras.preprocessing import image

from access_niu import persist


class Component(object):

    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        self._is_active = (
            kwargs.get("active") if kwargs.get("active") is not None else True
        )

    @abc.abstractmethod
    def name(self):
        raise NotImplementedError()

    def prepare(self, **kwargs):
        pass

    @abc.abstractmethod
    def execute(self, **kwargs):
        return NotImplementedError()

    def persist(self, **kwargs):
        pass

    @property
    def is_active(self):
        return self._is_active

    @is_active.setter
    def is_active(self, value):
        self._is_active = value


class MobilenetV2ModelComponent(Component):
    def __init__(self, **kwargs):
        super(MobilenetV2ModelComponent, self).__init__(**kwargs)
        self.model = self._build_model(**kwargs)

    def name(self):
        return "mobilenet_v2"

    def _build_model(self, input_layer, pretrained_model, output_layer, misc):
        input_tensor = Input(
            shape=(
                input_layer.get("image_height"),
                input_layer.get("image_width"),
                input_layer.get("color_channels"),
            )
        )

        base_model = MobileNetV2(input_tensor=input_tensor, **pretrained_model)
        x = base_model.output
        x = GlobalAveragePooling2D()(x)
        op = Dense(**output_layer)(x)

        for layer in base_model.layers:
            layer.trainable = False

        model = Model(inputs=[input_tensor], outputs=[op])
        model.compile(
            optimizer=misc.get("optimizer"),
            loss=misc.get("loss"),
            metrics=misc.get("metrics"),
        )

        return model

    def execute(self, **kwargs):

        train_generator = kwargs.get("train_generator")
        n_samples = kwargs.get("n_train_samples")
        batch_size = kwargs.get("batch_size")
        epochs = kwargs.get("epochs")

        self.model.fit_generator(
            generator=train_generator,
            steps_per_epoch=n_samples // batch_size, #TODO: handle case when steps_per_epoch==0
            epochs=epochs
        )

    def persist(self, **kwargs):

        labels = kwargs.get("labels")
        persist.save_keras_model(kwargs.get("project").get("path"), self.model, labels)


class DataGeneratorComponent(Component):
    def __init__(self, **kwargs):
        super(DataGeneratorComponent, self).__init__(**kwargs)

    def prepare(self, **kwargs):
        gen = image.ImageDataGenerator(rescale=1.0 / 255)

        generator = gen.flow_from_directory(
            kwargs.get("data_dir"),
            target_size=(kwargs.get("image_height"), kwargs.get("image_width")),
            batch_size=kwargs.get("batch_size"),
            shuffle=True,
        )

        labels = {v: k for k, v in generator.class_indices.items()}

        return {
            kwargs.get("generator_name"): generator,
            "labels": labels,
            kwargs.get("num_sample_name"): generator.n,
        }

    def execute(self, **kwargs):
        pass


class ComponentManager(object):
    def __init__(self):
        self.components = defaultdict(Component)
        self._load_components()

    def get(self, name):
        return self.components.get(name)

    def _load_components(self):
        self.components["mobilenet"] = MobileNet
        self.components["mobilenet_v2"] = MobilenetV2ModelComponent
        self.components["data_generator"] = DataGeneratorComponent



import os
import argparse

from flask import Flask, request, jsonify
import yaml

from access_niu.app import NIUApp


def _create_parser():
    parser = argparse.ArgumentParser(description="access-niu parser")
    parser.add_argument("--project", type=str, required=True, help="Path to trained model.")

    return parser.parse_args()


flask_app = Flask(__name__)

@flask_app.route("/")
def status():
    return {"status": "OK"}


@flask_app.route("/parse", methods=["POST"])
def parse():
    # model = request.form["model"]
    data = request.files["data"]
    resp = niu_app.parse(data)
    return jsonify(resp)


@flask_app.route("/train", methods=["POST"])
def train():
    pass


if __name__ == "__main__":
    args = _create_parser()

    with open(os.path.join(args.project, 'template.yml')) as f:
        niu_app = NIUApp(yaml.safe_load(f))

    flask_app.run(host="localhost", port=8000, debug=True)




# inspired from: https://github.com/explosion/spaCy/blob/master/spacy/about.py

__title__ = "access_niu"
__version__ = "0.0.1"
__summary__ = "Production grade Natural Image Understanding"
__uri__ = "http://accessai.co"
__author__ = "ConvexHull Technology"
__email__ = "contact@accessai.co"
__license__ = "Apache 2.0"
__release__ = True



from PIL import Image
import numpy as np

from werkzeug.datastructures import FileStorage

from access_niu.infer import load_model, parse
from access_niu.train import Trainer


class NIUApp(object):
    def __init__(self, template):
        self.template = template
        self.model_dir = template.get("project").get("path")
        self.model, self.labels = load_model(self.model_dir)
        self.trainer = Trainer(template)
        # TODO: Load it at project level
        self.image_size = self._get_image_dimensions(template)

    def _get_image_dimensions(self, template):
        input_layer = template.get("pipeline")[0].get("input_layer")

        return input_layer.get("image_width"), input_layer.get("image_height")

    def parse(self, data: FileStorage) -> dict:
        """Handles the parse request.

        :param data: Image file to be parsed.
        :return: dic of the parsed response
        """
        img = Image.open(data).resize(self.image_size, Image.ANTIALIAS)
        img = np.expand_dims(np.array(img), axis=0)
        pred_arr = parse(self.model, img)
        max_idx = np.argmax(pred_arr)
        return {self.labels.get(max_idx): np.float(pred_arr[0][max_idx])}

    def train(self):
        pass
